package main

import (
	"flag"
	"fmt"
	"hash/crc32"
	"log"
	"net/url"
	"os"
	"strings"

	"git.autistici.org/ale/autoradio"
	"github.com/gonuts/commander"
	gonutsflag "github.com/gonuts/flag"
)

type HasAddFlags interface {
	AddFlags(*gonutsflag.FlagSet)
}

type CommandInterface interface {
	Run([]string)
	Command() *commander.Command
}

type BaseCommand struct {
	UsageLine string
	Short     string
	Long      string
}

func (b *BaseCommand) Command() *commander.Command {
	return &commander.Command{
		UsageLine: b.UsageLine,
		Short:     b.Short,
		Long:      b.Long,
	}
}

func (b *BaseCommand) Run(args []string) {
}

func getClient() *autoradio.Client {
	return autoradio.NewClient(autoradio.NewEtcdClient(false))
}

func generateUsername(path string) string {
	return fmt.Sprintf("source%d", crc32.ChecksumIEEE([]byte(path)))
}

func setRelay(m *autoradio.Mount, relayUrl string) {
	if relayUrl == "" {
		// Randomly generate source credentials.
		m.Username = generateUsername(m.Name)
		m.Password = autoradio.GeneratePassword()
	} else {
		// Validate the given relay URL.
		u, err := url.Parse(relayUrl)
		if err != nil {
			log.Fatal(err)
		}
		m.RelayUrl = u.String()
		m.Username = ""
		m.Password = ""
	}
}

// Edit a mountpoint.
type editMountCommand struct {
	BaseCommand
	relay    string
	fallback string
}

var UNSET = "UNSET"

func newEditMountCommand() *editMountCommand {
	return &editMountCommand{
		BaseCommand: BaseCommand{
			UsageLine: "edit-mount <path>",
			Short:     "Edit an existing mountpoint",
			Long: `
Modify parameters of the specified mountpoint, such as the relay
and the fallback URL. If the relay option is set, the mountpoint
will not accept source connections anymore. To revert to the
default, non-relay behavior, set the relay to the empty string
(with --relay="").
`,
		},
	}
}

func (cmd *editMountCommand) AddFlags(f *gonutsflag.FlagSet) {
	// Note that we use special values to figure out whether a
	// flag has been specified or not. There might be better way
	// to do this.
	f.StringVar(&cmd.relay, "relay", UNSET, "Upstream URL to relay")
	f.StringVar(&cmd.fallback, "fallback", UNSET, "Fallback stream URL")
}

func (cmd *editMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}

	client := getClient()
	mount, err := client.GetMount(args[0])
	if err != nil {
		log.Fatal(err)
	}
	if mount == nil {
		log.Fatal("Mount not found")
	}

	if cmd.fallback != UNSET {
		mount.Fallback = cmd.fallback
	}
	if cmd.relay != UNSET {
		setRelay(mount, cmd.relay)
	}

	if err := client.SetMount(mount); err != nil {
		log.Fatal(err)
	}

	fmt.Printf("%+v\n", mount)
}

// Create a new mountpoint.
type createMountCommand struct {
	BaseCommand
	relay    string
	fallback string
}

func newCreateMountCommand() *createMountCommand {
	return &createMountCommand{
		BaseCommand: BaseCommand{
			UsageLine: "create-mount <path>",
			Short:     "Create a new mountpoint",
			Long: `
Create a new mountpoint at the specified PATH (which will become the
absolute URL of the stream).
`,
		},
	}
}

func (cmd *createMountCommand) AddFlags(f *gonutsflag.FlagSet) {
	f.StringVar(&cmd.relay, "relay", "", "Upstream URL to relay")
	f.StringVar(&cmd.fallback, "fallback", "", "Fallback stream URL")
}

func (cmd *createMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}
	fmt.Printf("args: %v\n", args)

	path := args[0]
	if !strings.HasPrefix(path, "/") {
		path = "/" + path
	}

	// Check if the mount already exists.
	client := getClient()
	if oldm, _ := client.GetMount(path); oldm != nil {
		log.Fatal("A mount with that name already exists!")
	}

	// Create the new mount and set the relevant fields (depending
	// on the options passed to the command).
	m := &autoradio.Mount{Name: path}
	setRelay(m, cmd.relay)
	if cmd.fallback != "" {
		m.Fallback = cmd.fallback
	}

	if err := client.SetMount(m); err != nil {
		log.Fatal(err)
	}

	fmt.Printf("%+v\n", m)
}

// Delete an existing mountpoint.
type deleteMountCommand struct {
	BaseCommand
}

func newDeleteMountCommand() *deleteMountCommand {
	return &deleteMountCommand{
		BaseCommand{
			UsageLine: "delete-mount <path>",
			Short:     "Delete a mountpoint",
			Long: `
Delete the specified mountpoint.
`,
		},
	}
}

func (cmd *deleteMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong number of arguments")
	}
	if err := getClient().DelMount(args[0]); err != nil {
		log.Fatal(err)
	}
	log.Printf("mountpoint %s removed", args[0])
}

// List known mountpoints.
type listMountsCommand struct {
	BaseCommand
}

func newListMountsCommand() *listMountsCommand {
	return &listMountsCommand{
		BaseCommand{
			UsageLine: "list-mounts",
			Short:     "List all configured mountpoints",
			Long: `
Outputs a list of all the currently configured mountpoints.
`,
		},
	}
}

func (cmd *listMountsCommand) Run(args []string) {
	if len(args) != 0 {
		log.Fatal("Too many arguments")
	}

	mounts, err := getClient().ListMounts()
	if err != nil {
		log.Fatal(err)
	}
	for _, m := range mounts {
		fmt.Println(m.Name)
	}
}

// Show mountpoint information.
type showMountCommand struct {
	BaseCommand
}

func newShowMountCommand() *showMountCommand {
	return &showMountCommand{
		BaseCommand{
			UsageLine: "show-mount <path>",
			Short:     "Show mountpoint information",
			Long: `
Print information about the specified mountpoint (including
the source credentials).
`,
		},
	}
}

func (cmd *showMountCommand) Run(args []string) {
	if len(args) != 1 {
		log.Fatal("Wrong nubmer of arguments")
	}
	mount, err := getClient().GetMount(args[0])
	if err != nil {
		log.Fatal(err)
	}
	if mount == nil {
		log.Fatal("Mount not found")
	}

	// Print out simple key=value pairs.
	fmt.Printf("path = %s\n", mount.Name)
	if mount.Username != "" {
		fmt.Printf("username = %s\n", mount.Username)
		fmt.Printf("password = %s\n", mount.Password)
	}
	if mount.RelayUrl != "" {
		fmt.Printf("relay_url = %s\n", mount.RelayUrl)
	}
	if mount.Fallback != "" {
		fmt.Printf("fallback_url = %s\n", mount.Fallback)
	}
}

var cmdr = &commander.Command{
	UsageLine: "radioctl <command> [<args>...]",
	Short:     "Manage `autoradio' configuration",
}

func addCommand(c CommandInterface) {
	cmd := c.Command()
	cmd.Run = func(unused *commander.Command, args []string) error {
		c.Run(args)
		return nil
	}
	if afc, ok := c.(HasAddFlags); ok {
		afc.AddFlags(&cmd.Flag)
	}
	cmdr.Subcommands = append(cmdr.Subcommands, cmd)
}

func init() {
	addCommand(newCreateMountCommand())
	addCommand(newEditMountCommand())
	addCommand(newDeleteMountCommand())
	addCommand(newListMountsCommand())
	addCommand(newShowMountCommand())
	flag.Usage = usage
}

func usage() {
	fmt.Fprintf(os.Stderr, "Usage: %s\n\nGlobal options:\n", cmdr.UsageLine)
	flag.PrintDefaults()
	fmt.Fprintln(os.Stderr, "\nType \"radioctl help\" for more documentation.\n")
}

func main() {
	flag.Parse()
	log.SetFlags(0)
	if err := cmdr.Dispatch(flag.Args()); err != nil {
		log.Fatal(err)
	}
}
